# -*- coding: utf-8; -*-
################################################################################
#
#  wuttaweb -- Web App for Wutta Framework
#  Copyright © 2024 Lance Edgar
#
#  This file is part of Wutta Framework.
#
#  Wutta Framework is free software: you can redistribute it and/or modify it
#  under the terms of the GNU General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option) any
#  later version.
#
#  Wutta Framework is distributed in the hope that it will be useful, but
#  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
#  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
#  more details.
#
#  You should have received a copy of the GNU General Public License along with
#  Wutta Framework.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Base grid classes
"""

import functools
import json
import logging
from collections import namedtuple

import sqlalchemy as sa
from sqlalchemy import orm

import paginate
from paginate_sqlalchemy import SqlalchemyOrmPage
from pyramid.renderers import render
from webhelpers2.html import HTML

from wuttaweb.db import Session
from wuttaweb.util import FieldList, get_model_fields, make_json_safe


log = logging.getLogger(__name__)


SortInfo = namedtuple('SortInfo', ['sortkey', 'sortdir'])
SortInfo.__doc__ = """
Named tuple to track sorting info.

Elements of :attr:`~Grid.sort_defaults` will be of this type.
"""

class Grid:
    """
    Base class for all grids.

    :param request: Reference to current :term:`request` object.

    :param columns: List of column names for the grid.  This is
       optional; if not specified an attempt will be made to deduce
       the list automatically.  See also :attr:`columns`.

    .. note::

       Some parameters are not explicitly described above.  However
       their corresponding attributes are described below.

    Grid instances contain the following attributes:

    .. attribute:: key

       Presumably unique key for the grid; used to track per-grid
       sort/filter settings etc.

    .. attribute:: vue_tagname

       String name for Vue component tag.  By default this is
       ``'wutta-grid'``.  See also :meth:`render_vue_tag()`.

    .. attribute:: model_class

       Model class for the grid, if applicable.  When set, this is
       usually a SQLAlchemy mapped class.  This may be used for
       deriving the default :attr:`columns` among other things.

    .. attribute:: columns

       :class:`~wuttaweb.util.FieldList` instance containing string
       column names for the grid.  Columns will appear in the same
       order as they are in this list.

       See also :meth:`set_columns()` and :meth:`get_columns()`.

    .. attribute:: data

       Data set for the grid.  This should either be a list of dicts
       (or objects with dict-like access to fields, corresponding to
       model records) or else an object capable of producing such a
       list, e.g. SQLAlchemy query.

       This is the "full" data set; see also
       :meth:`get_visible_data()`.

    .. attribute:: labels

       Dict of column label overrides.

       See also :meth:`get_label()` and :meth:`set_label()`.

    .. attribute:: renderers

       Dict of column (cell) value renderer overrides.

       See also :meth:`set_renderer()`.

    .. attribute:: actions

       List of :class:`GridAction` instances represenging action links
       to be shown for each record in the grid.

    .. attribute:: linked_columns

       List of column names for which auto-link behavior should be
       applied.

       See also :meth:`set_link()` and :meth:`is_linked()`.

    .. attribute:: sortable

       Boolean indicating whether *any* column sorting is allowed for
       the grid.  Default is ``False``.

       See also :attr:`sort_multiple` and :attr:`sort_on_backend`.

    .. attribute:: sort_multiple

       Boolean indicating whether "multi-column" sorting is allowed.
       Default is ``True``; if this is ``False`` then only one column
       may be sorted at a time.

       Only relevant if :attr:`sortable` is true, but applies to both
       frontend and backend sorting.

       .. warning::

          This feature is limited by frontend JS capabilities,
          regardless of :attr:`sort_on_backend` value (i.e. for both
          frontend and backend sorting).

          In particular, if the app theme templates use Vue 2 + Buefy,
          then multi-column sorting should work.

          But not so with Vue 3 + Oruga, *yet* - see also the `open
          issue <https://github.com/oruga-ui/oruga/issues/962>`_
          regarding that.  For now this flag is simply ignored for
          Vue 3 + Oruga templates.

          Additionally, even with Vue 2 + Buefy this flag can only
          allow the user to *request* a multi-column sort.  Whereas
          the "default sort" in the Vue component can only ever be
          single-column, regardless of :attr:`sort_defaults`.

    .. attribute:: sort_on_backend

       Boolean indicating whether the grid data should be sorted on the
       backend.  Default is ``True``.

       If ``False``, the client-side Vue component will handle the
       sorting.

       Only relevant if :attr:`sortable` is also true.

    .. attribute:: sorters

       Dict of functions to use for backend sorting.

       Only relevant if both :attr:`sortable` and
       :attr:`sort_on_backend` are true.

       See also :meth:`set_sorter()`, :attr:`sort_defaults` and
       :attr:`active_sorters`.

    .. attribute:: sort_defaults

       List of options to be used for default sorting, until the user
       requests a different sorting method.

       This list usually contains either zero or one elements.  (More
       are allowed if :attr:`sort_multiple` is true, but see note
       below.)  Each list element is a :class:`SortInfo` tuple and
       must correspond to an entry in :attr:`sorters`.

       Used with both frontend and backend sorting.

       See also :meth:`set_sort_defaults()` and
       :attr:`active_sorters`.

       .. warning::

          While the grid logic is built to handle multi-column
          sorting, this feature is limited by frontend JS
          capabilities.

          Even if ``sort_defaults`` contains multiple entries
          (i.e. for multi-column sorting to be used "by default" for
          the grid), only the *first* entry (i.e. single-column
          sorting) will actually be used as the default for the Vue
          component.

          See also :attr:`sort_multiple` for more details.

    .. attribute:: active_sorters

       List of sorters currently in effect for the grid; used by
       :meth:`sort_data()`.

       Whereas :attr:`sorters` defines all "available" sorters, and
       :attr:`sort_defaults` defines the "default" sorters,
       ``active_sorters`` defines the "current/effective" sorters.

       This attribute is set by :meth:`load_settings()`; until that is
       called it will not exist.

       This is conceptually a "subset" of :attr:`sorters` although a
       different format is used here::

          grid.active_sorters = [
              {'key': 'name', 'dir': 'asc'},
              {'key': 'id', 'dir': 'asc'},
          ]

       The above is for example only; there is usually no reason to
       set this attribute directly.

       This list may contain multiple elements only if
       :attr:`sort_multiple` is true.  Otherewise it should always
       have either zero or one element.

    .. attribute:: paginated

       Boolean indicating whether the grid data should be paginated,
       i.e. split up into pages.  Default is ``False`` which means all
       data is shown at once.

       See also :attr:`pagesize` and :attr:`page`, and
       :attr:`paginate_on_backend`.

    .. attribute:: paginate_on_backend

       Boolean indicating whether the grid data should be paginated on
       the backend.  Default is ``True`` which means only one "page"
       of data is sent to the client-side component.

       If this is ``False``, the full set of grid data is sent for
       each request, and the client-side Vue component will handle the
       pagination.

       Only relevant if :attr:`paginated` is also true.

    .. attribute:: pagesize_options

       List of "page size" options for the grid.  See also
       :attr:`pagesize`.

       Only relevant if :attr:`paginated` is true.  If not specified,
       constructor will call :meth:`get_pagesize_options()` to get the
       value.

    .. attribute:: pagesize

       Number of records to show in a data page.  See also
       :attr:`pagesize_options` and :attr:`page`.

       Only relevant if :attr:`paginated` is true.  If not specified,
       constructor will call :meth:`get_pagesize()` to get the value.

    .. attribute:: page

       The current page number (of data) to display in the grid.  See
       also :attr:`pagesize`.

       Only relevant if :attr:`paginated` is true.  If not specified,
       constructor will assume ``1`` (first page).
    """

    def __init__(
            self,
            request,
            vue_tagname='wutta-grid',
            model_class=None,
            key=None,
            columns=None,
            data=None,
            labels={},
            renderers={},
            actions=[],
            linked_columns=[],
            sortable=False,
            sort_multiple=True,
            sort_on_backend=True,
            sorters=None,
            sort_defaults=None,
            paginated=False,
            paginate_on_backend=True,
            pagesize_options=None,
            pagesize=None,
            page=1,
    ):
        self.request = request
        self.vue_tagname = vue_tagname
        self.model_class = model_class
        self.key = key
        self.data = data
        self.labels = labels or {}
        self.renderers = renderers or {}
        self.actions = actions or []
        self.linked_columns = linked_columns or []

        self.config = self.request.wutta_config
        self.app = self.config.get_app()

        self.set_columns(columns or self.get_columns())

        # sorting
        self.sortable = sortable
        self.sort_multiple = sort_multiple
        if self.sort_multiple and self.request.use_oruga:
            log.warning("grid.sort_multiple is not implemented for Oruga-based templates")
            self.sort_multiple = False
        self.sort_on_backend = sort_on_backend
        if sorters is not None:
            self.sorters = sorters
        elif self.sortable and self.sort_on_backend:
            self.sorters = self.make_backend_sorters()
        else:
            self.sorters = {}
        self.set_sort_defaults(sort_defaults or [])

        # paging
        self.paginated = paginated
        self.paginate_on_backend = paginate_on_backend
        self.pagesize_options = pagesize_options or self.get_pagesize_options()
        self.pagesize = pagesize or self.get_pagesize()
        self.page = page

    def get_columns(self):
        """
        Returns the official list of column names for the grid, or
        ``None``.

        If :attr:`columns` is set and non-empty, it is returned.

        Or, if :attr:`model_class` is set, the field list is derived
        from that, via :meth:`get_model_columns()`.

        Otherwise ``None`` is returned.
        """
        if hasattr(self, 'columns') and self.columns:
            return self.columns

        columns = self.get_model_columns()
        if columns:
            return columns

        return []

    def get_model_columns(self, model_class=None):
        """
        This method is a shortcut which calls
        :func:`~wuttaweb.util.get_model_fields()`.

        :param model_class: Optional model class for which to return
           fields.  If not set, the grid's :attr:`model_class` is
           assumed.
        """
        return get_model_fields(self.config,
                                model_class=model_class or self.model_class)

    @property
    def vue_component(self):
        """
        String name for the Vue component, e.g. ``'WuttaGrid'``.

        This is a generated value based on :attr:`vue_tagname`.
        """
        words = self.vue_tagname.split('-')
        return ''.join([word.capitalize() for word in words])

    def set_columns(self, columns):
        """
        Explicitly set the list of grid columns.

        This will overwrite :attr:`columns` with a new
        :class:`~wuttaweb.util.FieldList` instance.

        :param columns: List of string column names.
        """
        self.columns = FieldList(columns)

    def append(self, *keys):
        """
        Add some columns(s) to the grid.

        This is a convenience to allow adding multiple columns at
        once::

           grid.append('first_field',
                       'second_field',
                       'third_field')

        It will add each column to :attr:`columns`.
        """
        for key in keys:
            if key not in self.columns:
                self.columns.append(key)

    def remove(self, *keys):
        """
        Remove some column(s) from the grid.

        This is a convenience to allow removal of multiple columns at
        once::

           grid.remove('first_field',
                       'second_field',
                       'third_field')

        It will remove each column from :attr:`columns`.
        """
        for key in keys:
            if key in self.columns:
                self.columns.remove(key)

    def set_label(self, key, label):
        """
        Set/override the label for a column.

        :param key: Name of column.

        :param label: New label for the column header.

        See also :meth:`get_label()`.  Label overrides are tracked via
        :attr:`labels`.
        """
        self.labels[key] = label

    def get_label(self, key):
        """
        Returns the label text for a given column.

        If no override is defined, the label is derived from ``key``.

        See also :meth:`set_label()`.
        """
        if key in self.labels:
            return self.labels[key]
        return self.app.make_title(key)

    def set_renderer(self, key, renderer, **kwargs):
        """
        Set/override the value renderer for a column.

        :param key: Name of column.

        :param renderer: Callable as described below.

        Depending on the nature of grid data, sometimes a cell's
        "as-is" value will be undesirable for display purposes.

        The logic in :meth:`get_vue_data()` will first "convert" all
        grid data as necessary so that it is at least JSON-compatible.

        But then it also will invoke a renderer override (if defined)
        to obtain the "final" cell value.

        A renderer must be a callable which accepts 3 args ``(record,
        key, value)``:

        * ``record`` is the "original" record from :attr:`data`
        * ``key`` is the column name
        * ``value`` is the JSON-safe cell value

        Whatever the renderer returns, is then used as final cell
        value.  For instance::

           from webhelpers2.html import HTML

           def render_foo(record, key, value):
              return HTML.literal("<p>this is the final cell value</p>")

           grid = Grid(request, columns=['foo', 'bar'])
           grid.set_renderer('foo', render_foo)

        Renderer overrides are tracked via :attr:`renderers`.
        """
        if kwargs:
            renderer = functools.partial(renderer, **kwargs)
        self.renderers[key] = renderer

    def set_link(self, key, link=True):
        """
        Explicitly enable or disable auto-link behavior for a given
        column.

        If a column has auto-link enabled, then each of its cell
        contents will automatically be wrapped with a hyperlink.  The
        URL for this will be the same as for the "View"
        :class:`GridAction`
        (aka. :meth:`~wuttaweb.views.master.MasterView.view()`).
        Although of course each cell in the column gets a different
        link depending on which data record it points to.

        It is typical to enable auto-link for fields relating to ID,
        description etc. or some may prefer to auto-link all columns.

        See also :meth:`is_linked()`; the list is tracked via
        :attr:`linked_columns`.

        :param key: Column key as string.

        :param link: Boolean indicating whether column's cell contents
           should be auto-linked.
        """
        if link:
            if key not in self.linked_columns:
                self.linked_columns.append(key)
        else: # unlink
            if self.linked_columns and key in self.linked_columns:
                self.linked_columns.remove(key)

    def is_linked(self, key):
        """
        Returns boolean indicating if auto-link behavior is enabled
        for a given column.

        See also :meth:`set_link()` which describes auto-link behavior.

        :param key: Column key as string.
        """
        if self.linked_columns:
            if key in self.linked_columns:
                return True
        return False

    ##############################
    # sorting methods
    ##############################

    def make_backend_sorters(self, sorters=None):
        """
        Make backend sorters for all columns in the grid.

        This is called by the constructor, if both :attr:`sortable`
        and :attr:`sort_on_backend` are true.

        For each column in the grid, this checks the provided
        ``sorters`` and if the column is not yet in there, will call
        :meth:`make_sorter()` to add it.

        .. note::

           This only works if grid has a :attr:`model_class`.  If not,
           this method just returns the initial sorters (or empty
           dict).

        :param sorters: Optional dict of initial sorters.  Any
           existing sorters will be left intact, not replaced.

        :returns: Final dict of all sorters.  Includes any from the
           initial ``sorters`` param as well as any which were
           created.
        """
        sorters = sorters or {}

        if self.model_class:
            for key in self.columns:
                if key in sorters:
                    continue
                prop = getattr(self.model_class, key, None)
                if (prop and hasattr(prop, 'property')
                    and isinstance(prop.property, orm.ColumnProperty)):
                    sorters[prop.key] = self.make_sorter(prop)

        return sorters

    def make_sorter(self, columninfo, keyfunc=None, foldcase=True):
        """
        Returns a function suitable for use as a backend sorter on the
        given column.

        Code usually does not need to call this directly.  See also
        :meth:`set_sorter()`, which calls this method automatically.

        :param columninfo: Can be either a model property (see below),
           or a column name.

        :param keyfunc: Optional function to use as the "sort key
           getter" callable, if the sorter is manual (as opposed to
           SQLAlchemy query).  More on this below.  If not specified,
           a default function is used.

        :param foldcase: If the sorter is manual (not SQLAlchemy), and
           the column data is of text type, this may be used to
           automatically "fold case" for the sorting.  Defaults to
           ``True`` since this behavior is presumably expected, but
           may be disabled if needed.

        The term "model property" is a bit technical, an example
        should help to clarify::

           model = app.model
           grid = Grid(request, model_class=model.Person)

           # explicit property
           sorter = grid.make_sorter(model.Person.full_name)

           # property name works if grid has model class
           sorter = grid.make_sorter('full_name')

           # nb. this will *not* work
           person = model.Person(full_name="John Doe")
           sorter = grid.make_sorter(person.full_name)

        The ``keyfunc`` param allows you to override the way sort keys
        are obtained from data records (this only applies for a
        "manual" sort, where data is a list and not a SQLAlchemy
        query)::

           data = [
                {'foo': 1},
                {'bar': 2},
           ]

           # nb. no model_class, just as an example
           grid = Grid(request, columns=['foo', 'bar'], data=data)

           def getkey(obj):
               if obj.get('foo')
                   return obj['foo']
               if obj.get('bar'):
                   return obj['bar']
               return ''

           # nb. sortfunc will ostensibly sort by 'foo' column, but in
           # practice it is sorted per value from getkey() above
           sortfunc = grid.make_sorter('foo', keyfunc=getkey)
           sorted_data = sortfunc(data, 'asc')

        :returns: A function suitable for backend sorting.  This
           function will behave differently when it is given a
           SQLAlchemy query vs. a "list" of data.  In either case it
           will return the sorted result.

           This function may be called as shown above.  It expects 2
           args: ``(data, direction)``
        """
        model_class = None
        model_property = None
        if isinstance(columninfo, str):
            key = columninfo
            model_class = self.model_class
            model_property = getattr(self.model_class, key, None)
        else:
            model_property = columninfo
            model_class = model_property.class_
            key = model_property.key

        def sorter(data, direction):

            # query is sorted with order_by()
            if isinstance(data, orm.Query):
                if not model_property:
                    raise TypeError(f"grid sorter for '{key}' does not map to a model property")
                query = data
                return query.order_by(getattr(model_property, direction)())

            # other data is sorted manually.  first step is to
            # identify the function used to produce a sort key for
            # each record
            kfunc = keyfunc
            if not kfunc:
                if model_property:
                    # TODO: may need this for String etc. as well?
                    if isinstance(model_property.type, sa.Text):
                        if foldcase:
                            kfunc = lambda obj: (obj[key] or '').lower()
                        else:
                            kfunc = lambda obj: obj[key] or ''
                if not kfunc:
                    # nb. sorting with this can raise error if data
                    # contains varying types, e.g. str and None
                    kfunc = lambda obj: obj[key]

            # then sort the data and return
            return sorted(data, key=kfunc, reverse=direction == 'desc')

        # TODO: this should be improved; is needed in tailbone for
        # multi-column sorting with sqlalchemy queries
        if model_property:
            sorter._class = model_class
            sorter._column = model_property

        return sorter

    def set_sorter(self, key, sortinfo=None):
        """
        Set/override the backend sorter for a column.

        Only relevant if both :attr:`sortable` and
        :attr:`sort_on_backend` are true.

        :param key: Name of column.

        :param sortinfo: Can be either a sorter callable, or else a
           model property (see below).

        If ``sortinfo`` is a callable, it will be used as-is for the
        backend sorter.

        Otherwise :meth:`make_sorter()` will be called to obtain the
        backend sorter.  The ``sortinfo`` will be passed along to that
        call; if it is empty then ``key`` will be used instead.

        A backend sorter callable must accept ``(data, direction)``
        args and return the sorted data/query, for example::

           model = app.model
           grid = Grid(request, model_class=model.Person)

           def sort_full_name(query, direction):
              sortspec = getattr(model.Person.full_name, direction)
              return query.order_by(sortspec())

           grid.set_sorter('full_name', sort_full_name)

        See also :meth:`remove_sorter()` and :meth:`is_sortable()`.
        Backend sorters are tracked via :attr:`sorters`.
        """
        sorter = None

        if sortinfo and callable(sortinfo):
            sorter = sortinfo
        else:
            sorter = self.make_sorter(sortinfo or key)

        self.sorters[key] = sorter

    def remove_sorter(self, key):
        """
        Remove the backend sorter for a column.

        Note that this removes the sorter *function*, so there is
        no way to sort by this column unless another sorter is
        later defined for it.

        See also :meth:`set_sorter()`.
        """
        self.sorters.pop(key, None)

    def set_sort_defaults(self, *args):
        """
        Set the default sorting method for the grid.  This sorting is
        used unless/until the user requests a different sorting
        method.

        ``args`` for this method are interpreted as follows:

        If 2 args are received, they should be for ``sortkey`` and
        ``sortdir``; for instance::

           grid.set_sort_defaults('name', 'asc')

        If just one 2-tuple arg is received, it is handled similarly::

           grid.set_sort_defaults(('name', 'asc'))

        If just one string arg is received, the default ``sortdir`` is
        assumed::

           grid.set_sort_defaults('name') # assumes 'asc'

        Otherwise there should be just one list arg, elements of
        which are each 2-tuples of ``(sortkey, sortdir)`` info::

           grid.set_sort_defaults([('name', 'asc'),
                                   ('value', 'desc')])

        .. note::

           Note that :attr:`sort_multiple` determines whether the grid
           is actually allowed to have multiple sort defaults.  The
           defaults requested by the method call may be pruned if
           necessary to accommodate that.

        Default sorting info is tracked via :attr:`sort_defaults`.
        """

        # convert args to sort defaults
        sort_defaults = []
        if len(args) == 1:
            if isinstance(args[0], str):
                sort_defaults = [SortInfo(args[0], 'asc')]
            elif isinstance(args[0], tuple) and len(args[0]) == 2:
                sort_defaults = [SortInfo(*args[0])]
            elif isinstance(args[0], list):
                sort_defaults = [SortInfo(*tup) for tup in args[0]]
            else:
                raise ValueError("for just one positional arg, must pass string, 2-tuple or list")
        elif len(args) == 2:
            sort_defaults = [SortInfo(*args)]
        else:
            raise ValueError("must pass just one or two positional args")

        # prune if multi-column requested but not supported
        if len(sort_defaults) > 1 and not self.sort_multiple:
            log.warning("multi-column sorting is not enabled for the instance; "
                        "list will be pruned to first element for '%s' grid: %s",
                        self.key, sort_defaults)
            sort_defaults = [sort_defaults[0]]

        self.sort_defaults = sort_defaults

    def is_sortable(self, key):
        """
        Returns boolean indicating if a given column should allow
        sorting.

        If :attr:`sortable` is false, this always returns ``False``.

        For frontend sorting (i.e. :attr:`sort_on_backend` is false),
        this always returns ``True``.

        For backend sorting, may return true or false depending on
        whether the column is listed in :attr:`sorters`.

        :param key: Column key as string.

        See also :meth:`set_sorter()`.
        """
        if not self.sortable:
            return False
        if self.sort_on_backend:
            return key in self.sorters
        return True

    ##############################
    # paging methods
    ##############################

    def get_pagesize_options(self, default=None):
        """
        Returns a list of default page size options for the grid.

        It will check config but if no setting exists, will fall
        back to::

           [5, 10, 20, 50, 100, 200]

        :param default: Alternate default value to return if none is
           configured.

        This method is intended for use in the constructor.  Code can
        instead access :attr:`pagesize_options` directly.
        """
        options = self.config.get_list('wuttaweb.grids.default_pagesize_options')
        if options:
            options = [int(size) for size in options
                       if size.isdigit()]
            if options:
                return options

        return default or [5, 10, 20, 50, 100, 200]

    def get_pagesize(self, default=None):
        """
        Returns the default page size for the grid.

        It will check config but if no setting exists, will fall back
        to a value from :attr:`pagesize_options` (will return ``20`` if
        that is listed; otherwise the "first" option).

        :param default: Alternate default value to return if none is
           configured.

        This method is intended for use in the constructor.  Code can
        instead access :attr:`pagesize` directly.
        """
        size = self.config.get_int('wuttaweb.grids.default_pagesize')
        if size:
            return size

        if default:
            return default

        if 20 in self.pagesize_options:
            return 20

        return self.pagesize_options[0]

    ##############################
    # configuration methods
    ##############################

    def load_settings(self, persist=True):
        """
        Load all effective settings for the grid.

        If the request GET params (query string) contains grid
        settings, they are used; otherwise the settings are loaded
        from user session.

        .. note::

           As of now, "sorting" and "pagination" settings are the only
           type supported by this logic.  Settings for "filtering"
           coming soon...

        The overall logic for this method is as follows:

        * collect settings
        * apply settings to current grid
        * optionally save settings to user session

        Saving the settings to user session will allow the grid to
        remember its current settings when user refreshes the page, or
        navigates away then comes back.  Therefore normally, settings
        are saved each time they are loaded.  Note that such settings
        are wiped upon user logout.

        :param persist: Whether the collected settings should be saved
           to the user session.
        """

        # initial default settings
        settings = {}
        if self.sortable:
            if self.sort_defaults:
                # nb. as of writing neither Buefy nor Oruga support a
                # multi-column *default* sort; so just use first sorter
                sortinfo = self.sort_defaults[0]
                settings['sorters.length'] = 1
                settings['sorters.1.key'] = sortinfo.sortkey
                settings['sorters.1.dir'] = sortinfo.sortdir
            else:
                settings['sorters.length'] = 0
        if self.paginated and self.paginate_on_backend:
            settings['pagesize'] = self.pagesize
            settings['page'] = self.page

        # update settings dict based on what we find in the request
        # and/or user session.  always prioritize the former.

        if self.request_has_settings('sort'):
            self.update_sort_settings(settings, src='request')
            self.update_page_settings(settings)

        elif self.request_has_settings('page'):
            self.update_sort_settings(settings, src='session')
            self.update_page_settings(settings)

        else:
            # nothing found in request, so nothing new to save
            persist = False

            # but still should load whatever is in user session
            self.update_sort_settings(settings, src='session')
            self.update_page_settings(settings)

        # maybe save settings in user session, for next time
        if persist:
            self.persist_settings(settings, dest='session')

        # update ourself to reflect settings dict..

        # sorting
        if self.sortable:
            # nb. doing this for frontend sorting also
            self.active_sorters = []
            for i in range(1, settings['sorters.length'] + 1):
                self.active_sorters.append({
                    'key': settings[f'sorters.{i}.key'],
                    'dir': settings[f'sorters.{i}.dir'],
                })
                # TODO: i thought this was needed, but now idk?
                # # nb. when showing full index page (i.e. not partial)
                # # this implies we must set the default sorter for Vue
                # # component, and only single-column is allowed there.
                # if not self.request.GET.get('partial'):
                #     break

        # paging
        if self.paginated and self.paginate_on_backend:
            self.pagesize = settings['pagesize']
            self.page = settings['page']

    def request_has_settings(self, typ):
        """ """

        if typ == 'sort':
            if 'sort1key' in self.request.GET:
                return True

        elif typ == 'page':
            for key in ['pagesize', 'page']:
                if key in self.request.GET:
                    return True

        return False

    def get_setting(self, settings, key, src='session', default=None,
                    normalize=lambda v: v):
        """ """

        if src == 'request':
            value = self.request.GET.get(key)
            if value is not None:
                try:
                    return normalize(value)
                except ValueError:
                    pass

        elif src == 'session':
            value = self.request.session.get(f'grid.{self.key}.{key}')
            if value is not None:
                return normalize(value)

        # if src had nothing, try default/existing settings
        value = settings.get(key)
        if value is not None:
            return normalize(value)

        # okay then, default it is
        return default

    def update_sort_settings(self, settings, src=None):
        """ """
        if not (self.sortable and self.sort_on_backend):
            return

        if src == 'request':
            i = 1
            while True:
                skey = f'sort{i}key'
                if skey in self.request.GET:
                    settings[f'sorters.{i}.key'] = self.get_setting(settings, skey,
                                                                    src='request')
                    settings[f'sorters.{i}.dir'] = self.get_setting(settings, f'sort{i}dir',
                                                                    src='request',
                                                                    default='asc')
                else:
                    break
                i += 1
            settings['sorters.length'] = i - 1

        elif src == 'session':
            settings['sorters.length'] = self.get_setting(settings, 'sorters.length',
                                                          src='session', normalize=int)
            for i in range(1, settings['sorters.length'] + 1):
                for key in ('key', 'dir'):
                    skey = f'sorters.{i}.{key}'
                    settings[skey] = self.get_setting(settings, skey, src='session')

    def update_page_settings(self, settings):
        """ """
        if not (self.paginated and self.paginate_on_backend):
            return

        # update the settings dict from request and/or user session

        # pagesize
        pagesize = self.request.GET.get('pagesize')
        if pagesize is not None:
            if pagesize.isdigit():
                settings['pagesize'] = int(pagesize)
        else:
            pagesize = self.request.session.get(f'grid.{self.key}.pagesize')
            if pagesize is not None:
                settings['pagesize'] = pagesize

        # page
        page = self.request.GET.get('page')
        if page is not None:
            if page.isdigit():
                settings['page'] = int(page)
        else:
            page = self.request.session.get(f'grid.{self.key}.page')
            if page is not None:
                settings['page'] = int(page)

    def persist_settings(self, settings, dest=None):
        """ """
        if dest not in ('session',):
            raise ValueError(f"invalid dest identifier: {dest}")

        # func to save a setting value to user session
        def persist(key, value=lambda k: settings.get(k)):
            assert dest == 'session'
            skey = f'grid.{self.key}.{key}'
            self.request.session[skey] = value(key)

        # sort settings
        if self.sortable:

            # first must clear all sort settings from dest. this is
            # because number of sort settings will vary, so we delete
            # all and then write all

            if dest == 'session':
                # remove sort settings from user session
                prefix = f'grid.{self.key}.sorters.'
                for key in list(self.request.session):
                    if key.startswith(prefix):
                        del self.request.session[key]

            # now save sort settings to dest
            if 'sorters.length' in settings:
                persist('sorters.length')
                for i in range(1, settings['sorters.length'] + 1):
                    persist(f'sorters.{i}.key')
                    persist(f'sorters.{i}.dir')

        # pagination settings
        if self.paginated and self.paginate_on_backend:

            # save to dest
            persist('pagesize')
            persist('page')

    ##############################
    # data methods
    ##############################

    def get_visible_data(self):
        """
        Returns the "effective" visible data for the grid.

        This uses :attr:`data` as the starting point but may morph it
        for pagination etc. per the grid settings.

        Code can either access :attr:`data` directly, or call this
        method to get only the data for current view (e.g. assuming
        pagination is used), depending on the need.

        See also these methods which may be called by this one:

        * :meth:`sort_data()`
        * :meth:`paginate_data()`
        """
        data = self.data or []

        if self.sortable and self.sort_on_backend:
            data = self.sort_data(data)

        if self.paginated and self.paginate_on_backend:
            self.pager = self.paginate_data(data)
            data = self.pager

        return data

    def sort_data(self, data, sorters=None):
        """
        Sort the given data and return the result.  This is called by
        :meth:`get_visible_data()`.

        :param sorters: Optional list of sorters to use.  If not
           specified, the grid's :attr:`active_sorters` are used.
        """
        if sorters is None:
            sorters = self.active_sorters
        if not sorters:
            return data

        # nb. when data is a query, we want to apply sorters in the
        # requested order, so the final query has order_by() in the
        # correct "as-is" sequence.  however when data is a list we
        # must do the opposite, applying in the reverse order, so the
        # final list has the most "important" sort(s) applied last.
        if not isinstance(data, orm.Query):
            sorters = reversed(sorters)

        for sorter in sorters:
            sortkey = sorter['key']
            sortdir = sorter['dir']

            # cannot sort unless we have a sorter callable
            sortfunc = self.sorters.get(sortkey)
            if not sortfunc:
                return data

            # invoke the sorter
            data = sortfunc(data, sortdir)

        return data

    def paginate_data(self, data):
        """
        Apply pagination to the given data set, based on grid settings.

        This returns a "pager" object which can then be used as a
        "data replacement" in subsequent logic.

        This method is called by :meth:`get_visible_data()`.
        """
        if isinstance(data, orm.Query):
            pager = SqlalchemyOrmPage(data,
                                      items_per_page=self.pagesize,
                                      page=self.page)

        else:
            pager = paginate.Page(data,
                                  items_per_page=self.pagesize,
                                  page=self.page)

        # pager may have detected that our current page is outside the
        # valid range.  if so we should update ourself to match
        if pager.page != self.page:
            self.page = pager.page
            key = f'grid.{self.key}.page'
            if key in self.request.session:
                self.request.session[key] = self.page

            # and re-make the pager just to be safe (?)
            pager = self.paginate_data(data)

        return pager

    ##############################
    # rendering methods
    ##############################

    def render_vue_tag(self, **kwargs):
        """
        Render the Vue component tag for the grid.

        By default this simply returns:

        .. code-block:: html

           <wutta-grid></wutta-grid>

        The actual output will depend on various grid attributes, in
        particular :attr:`vue_tagname`.
        """
        return HTML.tag(self.vue_tagname, **kwargs)

    def render_vue_template(
            self,
            template='/grids/vue_template.mako',
            **context):
        """
        Render the Vue template block for the grid.

        This returns something like:

        .. code-block:: none

           <script type="text/x-template" id="wutta-grid-template">
             <b-table>
               <!-- columns etc. -->
             </b-table>
           </script>

        .. todo::

           Why can't Sphinx render the above code block as 'html' ?

           It acts like it can't handle a ``<script>`` tag at all?

        Actual output will of course depend on grid attributes,
        :attr:`vue_tagname` and :attr:`columns` etc.

        :param template: Path to Mako template which is used to render
           the output.
        """
        context['grid'] = self
        context.setdefault('request', self.request)
        output = render(template, context)
        return HTML.literal(output)

    def render_vue_finalize(self):
        """
        Render the Vue "finalize" script for the grid.

        By default this simply returns:

        .. code-block:: html

           <script>
             WuttaGrid.data = function() { return WuttaGridData }
             Vue.component('wutta-grid', WuttaGrid)
           </script>

        The actual output may depend on various grid attributes, in
        particular :attr:`vue_tagname`.
        """
        set_data = f"{self.vue_component}.data = function() {{ return {self.vue_component}Data }}"
        make_component = f"Vue.component('{self.vue_tagname}', {self.vue_component})"
        return HTML.tag('script', c=['\n',
                                     HTML.literal(set_data),
                                     '\n',
                                     HTML.literal(make_component),
                                     '\n'])

    def get_vue_columns(self):
        """
        Returns a list of Vue-compatible column definitions.

        This uses :attr:`columns` as the basis; each definition
        returned will be a dict in this format::

           {
               'field': 'foo',
               'label': "Foo",
               'sortable': True,
           }

        See also :meth:`get_vue_data()`.
        """
        if not self.columns:
            raise ValueError(f"you must define columns for the grid! key = {self.key}")

        columns = []
        for name in self.columns:
            columns.append({
                'field': name,
                'label': self.get_label(name),
                'sortable': self.is_sortable(name),
            })
        return columns

    def get_vue_active_sorters(self):
        """
        Returns a list of Vue-compatible column sorter definitions.

        The list returned is the same as :attr:`active_sorters`;
        however the format used in Vue is different.  So this method
        just "converts" them to the required format, e.g.::

           # active_sorters format
           {'key': 'name', 'dir': 'asc'}

           # get_vue_active_sorters() format
           {'field': 'name', 'order': 'asc'}

        :returns: The :attr:`active_sorters` list, converted as
           described above.
        """
        sorters = []
        for sorter in self.active_sorters:
            sorters.append({'field': sorter['key'],
                            'order': sorter['dir']})
        return sorters

    def get_vue_data(self):
        """
        Returns a list of Vue-compatible data records.

        This calls :meth:`get_visible_data()` but then may modify the
        result, e.g. to add URLs for :attr:`actions` etc.

        Importantly, this also ensures each value in the dict is
        JSON-serializable, using
        :func:`~wuttaweb.util.make_json_safe()`.

        :returns: List of data record dicts for use with Vue table
           component.  May be the full set of data, or just the
           current page, per :attr:`paginate_on_backend`.
        """
        original_data = self.get_visible_data()

        # TODO: at some point i thought it was useful to wrangle the
        # columns here, but now i can't seem to figure out why..?

        # # determine which columns are relevant for data set
        # columns = None
        # if not columns:
        #     columns = self.get_columns()
        #     if not columns:
        #         raise ValueError("cannot determine columns for the grid")
        # columns = set(columns)
        # if self.model_class:
        #     mapper = sa.inspect(self.model_class)
        #     for column in mapper.primary_key:
        #         columns.add(column.key)

        # # prune data fields for which no column is defined
        # for i, record in enumerate(original_data):
        #     original_data[i]= dict([(key, record[key])
        #                             for key in columns])

        # we have action(s), so add URL(s) for each record in data
        data = []
        for i, record in enumerate(original_data):
            original_record = record

            record = dict(record)

            # convert data if needed, for json compat
            record = make_json_safe(record,
                                    # TODO: is this a good idea?
                                    warn=False)

            # customize value rendering where applicable
            for key in self.renderers:
                value = record.get(key, None)
                record[key] = self.renderers[key](original_record, key, value)

            # add action urls to each record
            for action in self.actions:
                key = f'_action_url_{action.key}'
                if key not in record:
                    url = action.get_url(original_record, i)
                    if url:
                        record[key] = url

            data.append(record)

        return data

    def get_vue_pager_stats(self):
        """
        Returns a simple dict with current grid pager stats.

        This is used when :attr:`paginate_on_backend` is in effect.
        """
        pager = self.pager
        return {
            'item_count': pager.item_count,
            'items_per_page': pager.items_per_page,
            'page': pager.page,
            'page_count': pager.page_count,
            'first_item': pager.first_item,
            'last_item': pager.last_item,
        }


class GridAction:
    """
    Represents a "row action" hyperlink within a grid context.

    All such actions are displayed as a group, in a dedicated
    **Actions** column in the grid.  So each row in the grid has its
    own set of action links.

    A :class:`Grid` can have one (or zero) or more of these in its
    :attr:`~Grid.actions` list.  You can call
    :meth:`~wuttaweb.views.base.View.make_grid_action()` to add custom
    actions from within a view.

    :param request: Current :term:`request` object.

    .. note::

       Some parameters are not explicitly described above.  However
       their corresponding attributes are described below.

    .. attribute:: key

       String key for the action (e.g. ``'edit'``), unique within the
       grid.

    .. attribute:: label

       Label to be displayed for the action link.  If not set, will be
       generated from :attr:`key` by calling
       :meth:`~wuttjamaican:wuttjamaican.app.AppHandler.make_title()`.

       See also :meth:`render_label()`.

    .. attribute:: url

       URL for the action link, if applicable.  This *can* be a simple
       string, however that will cause every row in the grid to have
       the same URL for this action.

       A better way is to specify a callable which can return a unique
       URL for each record.  The callable should expect ``(obj, i)``
       args, for instance::

          def myurl(obj, i):
              return request.route_url('widgets.view', uuid=obj.uuid)

          action = GridAction(request, 'view', url=myurl)

       See also :meth:`get_url()`.

    .. attribute:: icon

       Name of icon to be shown for the action link.

       See also :meth:`render_icon()`.

    .. attribute:: link_class

       Optional HTML class attribute for the action's ``<a>`` tag.
    """

    def __init__(
            self,
            request,
            key,
            label=None,
            url=None,
            icon=None,
            link_class=None,
    ):
        self.request = request
        self.config = self.request.wutta_config
        self.app = self.config.get_app()
        self.key = key
        self.url = url
        self.label = label or self.app.make_title(key)
        self.icon = icon or key
        self.link_class = link_class or ''

    def render_icon_and_label(self):
        """
        Render the HTML snippet for action link icon and label.

        Default logic returns the output from :meth:`render_icon()`
        and :meth:`render_label()`.
        """
        html = [
            self.render_icon(),
            self.render_label(),
        ]
        return HTML.literal(' ').join(html)

    def render_icon(self):
        """
        Render the HTML snippet for the action link icon.

        This uses :attr:`icon` to identify the named icon to be shown.
        Output is something like (here ``'trash'`` is the icon name):

        .. code-block:: html

           <i class="fas fa-trash"></i>

        See also :meth:`render_icon_and_label()`.
        """
        if self.request.use_oruga:
            return HTML.tag('o-icon', icon=self.icon)

        return HTML.tag('i', class_=f'fas fa-{self.icon}')

    def render_label(self):
        """
        Render the label text for the action link.

        Default behavior is to return :attr:`label` as-is.

        See also :meth:`render_icon_and_label()`.
        """
        return self.label

    def get_url(self, obj, i=None):
        """
        Returns the action link URL for the given object (model
        instance).

        If :attr:`url` is a simple string, it is returned as-is.

        But if :attr:`url` is a callable (which is typically the most
        useful), that will be called with the same ``(obj, i)`` args
        passed along.

        :param obj: Model instance of whatever type the parent grid is
           setup to use.

        :param i: Zero-based sequence for the object, within the
           parent grid.

        See also :attr:`url`.
        """
        if callable(self.url):
            return self.url(obj, i)

        return self.url
